/*
 *  2ooM: The Master of Orion II Reverse Engineering Project
 *  Simple command-line tool to extract LBX archive files.
 *  Copyright © 2006-2011, 2013 Nick Bowler
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <getopt.h>
#include <fnmatch.h>

#include "tools.h"
#include "lbx.h"
#include "error.h"

static void printusage(void)
{
	puts("usage: lbxtool [-l|-x] [-v] [-f path] [file ...]");
}

static void printhelp(void)
{
	printusage();
	puts("For now, see the man page for detailed help.");
}

enum {
	MODE_NONE,
	MODE_LIST,
	MODE_EXTRACT,
};

int filematch(char **argv, const char *name)
{
	int i;

	for (i = 0; argv[i]; i++) {
		switch(fnmatch(argv[i], name, 0)) {
		case 0:
			return 0;
		case FNM_NOMATCH:
			break;
		default:
			tool_err(-1, "error matching glob: %s", argv[i]);
			return 1;
		}
	}

	return i ? -1: 0;
}

int list(struct lbx *lbx, int verbose, char **argv)
{
	unsigned int i;

	if (verbose) {
		printf("Files in archive: %u\n", lbx->nfiles);
	}

	for (i = 0; i < lbx->nfiles; i++) {
		struct lbx_statbuf stat;

		lbx_file_stat(lbx, i, &stat);

		switch (filematch(argv, stat.name)) {
		case -1: continue;
		case  0: break;
		default: return EXIT_FAILURE;
		}

		printf("%s", stat.name);
		if (verbose) {
			printf(" size=%zu bytes", stat.size);
		}

		putchar('\n');
	}

	return EXIT_SUCCESS;
}

int extract_file(LBXfile *f, const struct lbx_statbuf *stat)
{
	int ret = -1;
	size_t rc;
	FILE *of;

	of = fopen(stat->name, "wb");
	if (!of) {
		tool_err(0, "%s: fopen", stat->name);
		return -1;
	}

	while (1) {
		unsigned char buf[1024];

		rc = lbx_file_read(f, buf, sizeof buf);
		if (rc == 0) {
			if (lbx_file_eof(f))
				ret = 0;
			break;
		}

		if (fwrite(buf, rc, 1, of) != 1) {
			tool_err(0, "%s: fwrite", stat->name);
			break;
		}

		if (rc < sizeof buf) {
			if (lbx_file_eof(f))
				ret = 0;
			break;
		}
	}

	if (fclose(of) == EOF) {
		tool_err(0, "%s: fclose", stat->name);
		return -1;
	}

	return ret;
}

int extract(struct lbx *lbx, int verbose, char **argv)
{
	unsigned int i;

	if (verbose) {
		printf("Files in archive: %u\n", lbx->nfiles);
	}

	for (i = 0; i < lbx->nfiles; i++) {
		struct lbx_statbuf stat;
		LBXfile *file;

		lbx_file_stat(lbx, i, &stat);

		switch (filematch(argv, stat.name)) {
		case -1: continue;
		case  0: break;
		default: return EXIT_FAILURE;
		}

		file = lbx_file_open(lbx, i);
		if (!file) {
			tool_err(-1, "%s: %s", stat.name, lbx_errmsg());
			continue;
		}

		if (extract_file(file, &stat) == -1)
			return EXIT_FAILURE;
		lbx_file_close(file);
	}

	return EXIT_SUCCESS;
}

int main(int argc, char **argv)
{
	int mode = MODE_NONE, verbose = 0, opt, rc = EXIT_FAILURE;
	struct lbx_pipe_state stdin_handle = { .f = stdin };
	const char *file = NULL;
	struct lbx *lbx;

	static const char         *sopts   = "lxf:i:vV";
	static const struct option lopts[] = {
		{ "list",    0, NULL, 'l' },
		{ "extract", 0, NULL, 'x' },

		{ "file",    1, NULL, 'f' },
		{ "index",   1, NULL, 'i' },

		{ "verbose", 0, NULL, 'v' },

		{ "version", 0, NULL, 'V' },
		{ "usage",   0, NULL, 'U' },
		{ "help",    0, NULL, 'H' },

		{ 0 }
	};

	tool_init("lbxtool", argc, argv);
	while ((opt = getopt_long(argc, argv, sopts, lopts, NULL)) != -1) {
		switch(opt) {
		case 'l':
			mode = MODE_LIST;
			break;
		case 'x':
			mode = MODE_EXTRACT;
			break;
		case 'f':
			file = optarg;
			break;
		case 'i':
			/* FIXME: Add index file support. */
			break;
		case 'v':
			verbose = 1;
			break;
		case 'V':
			tool_version();
			return EXIT_SUCCESS;
		case 'U':
			printusage();
			return EXIT_SUCCESS;
		case 'H':
			printhelp();
			return EXIT_SUCCESS;
		default:
			return EXIT_FAILURE;
		}
	}

	if (file)
		lbx = lbx_fopen(file);
	else
		lbx = lbx_open(&stdin_handle, &lbx_pipe_fops, NULL, "stdin");

	if (!lbx) {
		tool_err(-1, "%s: %s", file ? file : "stdin", lbx_errmsg());
		return EXIT_FAILURE;
	}

	switch (mode) {
	case MODE_LIST:
		rc = list(lbx, verbose, &argv[optind]);
		break;
	case MODE_EXTRACT:
		rc = extract(lbx, verbose, &argv[optind]);
		break;
	default:
		tool_err(-1, "no mode specified");
	}

	lbx_close(lbx);
	return rc;
}
